/*
 * - error handling (mkl: 2005/09/12)
 */
#include <errno.h>
#include <stdio.h>
#include <unistd.h>

#include <sys/types.h>
#include <sys/socket.h>

#include <gsloop_types.h>
#include <gsloop_server.h>
#include <gsloop_client.h>
#include <gsloop_con.h>


int gsloop_server_attach_client(struct gsloop	*gsloop,
				int		fd)
{
	int			fd_client;
	int			err;
	struct gsloop_client	*client;
	struct gsloop_server	*server;

	if (!gsloop)
		return -EINVAL;
	if (fd < 0)
		return -EBADF;


	fd_client = accept(fd, NULL, 0);
	if (fd_client < 0) {
		perror("accept");
		err = -errno;
		goto exit_0;
	}
				
	err = rn_prepare_fd_for_add(fd_client, getpid());
	if (err) {
		err = -err;	/* rn return errno*/
		goto exit_1;
	}
			
	err = rn_add(&gsloop->rn, fd_client, gsloop_client_cb, gsloop);
	if (err) {
		err = -err;	/* rn return errno */
		goto exit_2;
	}

	client = &gsloop_con_alloc_and_add(gsloop, fd_client)->client;
	if (!client) {
		err = -ENOMEM;
		goto exit_3;
	}

	server = gsloop_con_get_server(gsloop, fd);
	if (!server) {
		err = -ENOMEM;
		goto exit_4;
	}
	
	client->data = server->client_data;
	client->fd = fd_client;
	client->close = server->client_close;
	client->cb = server->client_cb;

	/* FIXME: check for max allowed connections */

	dprintf(LEVEL_INFO, "[fd=%02d] [new_fd=%02d] new connection from client\n", fd, fd_client);

	return fd_client;

 exit_4:
 exit_3:
	rn_del(&gsloop->rn, fd_client);
 exit_2:
 exit_1:
	close(fd_client);
 exit_0:
	return err;
}


int gsloop_server_callback(rn_pollevent_t	*event)
{
	struct gsloop	*gsloop = event->client.data;
	int		fd = event->fd;
	int		err = 0;

	if (event->revents & POLLIN) {
		err = gsloop_server_attach_client(gsloop, fd);

		/* FIXME: err = gsloop_server_attach_client_hook(gsl, fd); */

	} else {
		switch (event->revents) {
		case POLLOUT:
			printf("POLLOUT\n");
			break;
		default:
			printf("Event: %d\n", event->revents);
			break;
		}
	}

	return err;
}
